module Rev2D_fd_mod

  use Rev2D_fd_types

  use genrev3d_types
  use rev_extrap_types
  use data_types
  use slow_types
  use wave_types


  implicit none

  logical inject_data
  integer, private :: i_z_sou,i_z_rec
  type (Rev2D_fd_param_type), private :: Rev2D_fd_par_loc

contains

  subroutine init_Rev2D_fd_par(Rev2D_fd_par)
    type (Rev2D_fd_param_type) Rev2D_fd_par

    Rev2D_fd_par_loc=Rev2D_fd_par

  end subroutine init_Rev2D_fd_par

  function Rev2D_fd_init(gen_par,data_par,slow_par,wave_par) &
                                                        result(info)
  type (genrev3d_info_type) info
  type (genrev3d_param_type) gen_par
  type (data_param_type) data_par
  type (slow_param_type) slow_par
  type (wave_param_type) wave_par
  
  integer ierr,i_z,n_z,m_mx_pos


  info=info_default

  inject_data=gen_par%inject_data
    write(0,*)'inject_data=',inject_data

  wave_par%n_t_step=3

  i_z_sou=wave_par%i_z_sou
  i_z_rec=wave_par%i_z_rec

  if(mod(wave_par%n_mx,2) /=0) then
    write(0,*)'n_mx=',wave_par%n_mx
    call seperr ("n_mx must be even")
  end if  
  end function Rev2D_fd_init 



  function Rev2D_fd_extrap_oper(rev_extrap_par,wave_par,&
                          force_block,wave_block,slow_block)result(stat)
  integer stat
  type (rev_extrap_param_type) rev_extrap_par
  type (wave_param_type) wave_par

  integer omp_get_thread_num
  integer n_threads,i_thread
  
  logical add,adj
  integer n_z,n_mx,i_z,i_mx,ierr
  integer i_t0,i_tm1,i_tm2
  real vel_sq
  real d_t_sq

  real d_x_sq_reci,d_z_sq_reci
  real lap_tm1_x,lap_tm1_z,lap_tm1
  real c0,c1,c2
  real bndz0,delp,vel,d_tz,d_tx,bndx0,bndx1
  integer iop,nop,i_z_force

  optional slow_block
  real,  dimension (:,:,:,:)      :: force_block
  real,  dimension (:,:,:,:,:,:)  :: wave_block
  real,  dimension (:,:,:,:,:,:)  :: slow_block

  adj=rev_extrap_par%adj
  add=rev_extrap_par%add

  n_threads=rev_extrap_par%n_threads

  d_t_sq=(rev_extrap_par%d_t)**2
  d_x_sq_reci=1.0/((wave_par%d_mx)**2)
  d_z_sq_reci=1.0/((wave_par%d_z)**2)
  d_tz=rev_extrap_par%d_t/wave_par%d_z
  d_tx=rev_extrap_par%d_t/wave_par%d_mx

  i_t0=wave_par%i_t_step(1)
  i_tm1=wave_par%i_t_step(2)
  i_tm2=wave_par%i_t_step(3)

  select case (rev_extrap_par%sou_rec_wave)
   case (SOU_WAVE)
     i_z_force=i_z_sou
   case (REC_WAVE)
     i_z_force=i_z_rec
   case default
     write(0,*)'sou_rec_wave=',rev_extrap_par%sou_rec_wave
     call seperr('Rev2D_fd_mod: Not supported sou_rec_wave')
  end select

  n_z=wave_par%n_z
  n_mx=wave_par%n_mx
!  do i_z=1,n_z
!       do i_mx=1,n_mx  
!            slow_block(i_z,i_mx,1,1,1,1)=1.0/2000.0
!       end do
!  end do


  c0 = -30./12.
  c1 = +16./12.
  c2 = -1./12.
  nop=2
  
  !$omp parallel default (shared), &
  !$omp& private(i_mx,i_z)
  !$omp do
  do i_mx=1,n_mx
    do i_z=1,n_z
       wave_block(i_tm2,i_z,i_mx,1,1,1)= wave_block(i_tm1,i_z,i_mx,1,1,1)
       wave_block(i_tm1,i_z,i_mx,1,1,1)= wave_block(i_t0,i_z,i_mx,1,1,1)
    end do
  end do
!$omp end parallel

  !$omp parallel default (shared), &
  !$omp& private(i_mx,i_z,vel_sq,lap_tm1_x,lap_tm1_z,lap_tm1)
  !$omp do
  do i_mx=1+nop,n_mx-nop
!write(0,*)'Thred #',omp_get_thread_num()
!if(omp_get_thread_num() == 0) then
!write(0,*)'Thred 0'
!end if
!if(omp_get_thread_num() == 1) then
!write(0,*)'Thred 1'
!end if
    do i_z=1+nop,n_z-nop
      vel_sq=1.0/(slow_block(i_z,i_mx,1,1,1,1)**2)

      lap_tm1_x=c0*wave_block(i_tm1,i_z,i_mx,1,1,1)+ &
        c1*(wave_block(i_tm1,i_z,i_mx-1,1,1,1) &
                              +wave_block(i_tm1,i_z,i_mx+1,1,1,1))&
       +c2*(wave_block(i_tm1,i_z,i_mx-2,1,1,1) &
                              +wave_block(i_tm1,i_z,i_mx+2,1,1,1))
      lap_tm1_x=lap_tm1_x*d_x_sq_reci
      lap_tm1_z=c0*wave_block(i_tm1,i_z,i_mx,1,1,1)+ &
                c1*(wave_block(i_tm1,i_z-1,i_mx,1,1,1) &
                            +wave_block(i_tm1,i_z+1,i_mx,1,1,1) )+ &
                c2*(wave_block(i_tm1,i_z-2,i_mx,1,1,1) &
                              +wave_block(i_tm1,i_z+2,i_mx,1,1,1))
      lap_tm1_z=lap_tm1_z*d_z_sq_reci
      lap_tm1=lap_tm1_x+lap_tm1_z

      wave_block(i_t0,i_z,i_mx,1,1,1)= 2*wave_block(i_tm1,i_z,i_mx,1,1,1)&
      - wave_block(i_tm2,i_z,i_mx,1,1,1)&
      + d_t_sq*vel_sq*lap_tm1     
    end do
  end do
!$omp end parallel

!###################### absorbing boundary condition #######
! Not easy to parallelize
!  !$omp parallel default (shared), &
!  !$omp& private(iop,i_mx,i_z,vel,delp,bndz0,bndx0,bndx1)
!  !$omp do
  do iop=nop,1,-1
     do i_mx=1,n_mx
        vel=1.0/slow_block(iop,i_mx,1,1,1,1)
        delp=vel*d_tz
        bndz0= (1.-delp)/(1.+delp) 
        wave_block(i_t0,iop,i_mx,1,1,1)= &
                    wave_block(i_tm1,iop+1,i_mx,1,1,1)+ &
             bndz0*(wave_block(i_tm1,iop,i_mx,1,1,1)- &
              wave_block(i_t0,iop+1,i_mx,1,1,1) )  
     end do
     do i_mx=1,n_mx
        vel=1.0/slow_block(n_z-iop+1,i_mx,1,1,1,1)
        delp=vel*d_tz
        bndz0= (1.-delp)/(1.+delp) 
        wave_block(i_t0,n_z-iop+1,i_mx,1,1,1)= &
                    wave_block(i_tm1,n_z-iop,i_mx,1,1,1)+ &
             bndz0*(wave_block(i_tm1,n_z-iop+1,i_mx,1,1,1)- &
              wave_block(i_t0,n_z-iop,i_mx,1,1,1) )  
     end do

     
     do i_z=1,n_z
        vel=1.0/slow_block(i_z,iop,1,1,1,1)
        delp=vel*d_tx
        bndx0= (1.-delp)/(1.+delp)
        wave_block(i_t0,i_z,iop,1,1,1)=&
           wave_block(i_tm1,i_z,iop+1,1,1,1)+ &
           bndx0*(wave_block(i_tm1,i_z,iop,1,1,1)- &     
                wave_block(i_t0,i_z,iop+1,1,1,1) )     
     end do


     do i_z=1,n_z
        vel=1.0/slow_block(i_z,n_mx-iop+1,1,1,1,1)
        delp = vel*d_tx
        bndx1= (1.-delp)/(1.+delp) 
        wave_block(i_t0,i_z,n_mx-iop+1,1,1,1)=&
            wave_block(i_tm1,i_z,n_mx-iop,1,1,1)+ &   
            bndx1*(wave_block(i_tm1,i_z,n_mx-iop+1,1,1,1)- &
                  wave_block(i_t0,i_z,n_mx-iop,1,1,1) )  
     end do

  enddo
!!$omp end parallel

!####################### end absorbing boundary condition ####
  
!@@@@@@@@@@@@@@@@@@@@@@ source force @@@@@@@@@@@@@@@@@@@@@@@@@
  if(inject_data) then
  !$omp parallel default (shared), &
  !$omp& private(i_mx)
  !$omp do
    do i_mx=1,n_mx
      wave_block(i_t0,i_z_force,i_mx,1,1,1)= &
                  force_block(i_mx,1,1,1)
    end do
  !$omp end parallel
  else 
  !$omp parallel default (shared), &
  !$omp& private(i_mx)
  !$omp do
    do i_mx=1,n_mx
       wave_block(i_t0,i_z_force,i_mx,1,1,1)=&
           wave_block(i_t0,i_z_force,i_mx,1,1,1) &
           +d_t_sq*force_block(i_mx,1,1,1)
    end do
  !$omp end parallel
  end if

!@@@@@@@@@@@@@@@@@@@@@@ end source force@@@@@@@@@@@@@@@@@@@@@@
  stat=0
  return
  end function Rev2D_fd_extrap_oper

end module Rev2D_fd_mod
