module Rev2D_spectral_mod

  use Rev2D_spectral_types

  use genrev3d_types
  use rev_extrap_types
  use data_types
  use slow_types
  use wave_types
  use wavelet_types
  use fft_types
  use bound_types

  use fft_data_mod

  use bound_mod

  implicit none

  logical inject_data,limit_k
  integer, private :: i_z_sou,i_z_rec,half_n_z_sou,half_n_z_rec
  type (Rev2D_spectral_param_type), private :: Rev2D_spectral_par_loc

  logical, private :: apply_any_bound

  integer, private :: n_k_mx_pos,n_k_mx_max
! n_k_mx_pos - number of positive wavenumber along mx

  real, allocatable, dimension (:), private :: k_mx,k_z
  real, allocatable, dimension (:,:), private :: ksq
! k_mx - wavenumbers along mx
! k_z - wavenumbers along z
  complex, allocatable, dimension (:,:), private :: temp,temp_mx,temp_z
! temp - temporary array 
! temp_mx - temporary vector 
! temp_z - temporary vector 
  type (fft_type), dimension(:), pointer, private :: fft_temp_z_id,fft_temp_mx_id
! fft_temp_z_id - fft data structure fro 1-D FFT along mx
! fft_temp_mx_id - fft data structure fro 1-D FFT along z


contains

  subroutine init_Rev2D_spectral_par(Rev2D_spectral_par)
    type (Rev2D_spectral_param_type) Rev2D_spectral_par

! Copy outside parameter structure into local structure
    Rev2D_spectral_par_loc=Rev2D_spectral_par

  end subroutine init_Rev2D_spectral_par

! Initialize local variables
! Sallocate and initialize local array
  function Rev2D_spectral_init(gen_par,data_par,slow_par, &
		wave_par,wavelet_par,bound_par) &
          result(info)
    type (genrev3d_info_type) info
    type (genrev3d_param_type) gen_par
    type (data_param_type) data_par
    type (slow_param_type) slow_par
    type (wave_param_type) wave_par
    type (wavelet_param_type) wavelet_par
    type (bound_param_type) bound_par

    integer i_k_mx,i_k_hx,i_k_z,ierr,i_z,n_z,m_mx_pos,i_thread
    real kz,kmx,khx,epsi,k_z_max,k_mx_max,k_max,kk_z,kk_mx

    info=info_default


    gen_par%max_courant_numb=sqrt(2.)/acos(-1.)

    apply_any_bound=gen_par%apply_bound

    inject_data=gen_par%inject_data
    if(gen_par%debug1) write(0,*)'inject_data=',inject_data

    wave_par%n_t_step=3

    i_z_sou=wave_par%i_z_sou
    i_z_rec=wave_par%i_z_rec
    half_n_z_sou=(wavelet_par%n_z-1)/2
    half_n_z_rec=0 ! ATTENTION if this changes data_mod.f90 must change too!

    if(gen_par%debug1) write(0,*)'SOU_WAVE: i_z_sou=',i_z_sou
    if(gen_par%debug1) write(0,*)'SOU_WAVE: half_n_z_sou=',half_n_z_sou
    if(gen_par%debug1) write(0,*)'REC_WAVE: i_z_rec=',i_z_rec
    if(gen_par%debug1) write(0,*)'REC_WAVE: half_n_z_rec=',half_n_z_rec
    if(gen_par%debug1) write(0,*)'bound nz=',bound_par%n_z_bound

    if(i_z_sou-half_n_z_sou < bound_par%n_z_bound+1) then
      call seperr('STOP: you are placing source in top absorbing boundary')
    end if
    if (i_z_sou+half_n_z_sou > wave_par%n_z-bound_par%n_z_bound) then
      call seperr('STOP: you are placing source in bottom absorbing boundary')
    end if

    if(i_z_rec-half_n_z_rec < bound_par%n_z_bound+1) then
      call seperr('STOP: you are placing receivers in top absorbing boundary')
    end if
    if (i_z_rec+half_n_z_rec > wave_par%n_z-bound_par%n_z_bound) then
     call seperr('STOP: you are placing receivers in bottom absorbing boundary')
    end if


    if(mod(wave_par%n_mx,2) /=0) then
      write(0,*)'n_mx=',wave_par%n_mx
      call seperr ("n_mx must be even")
    end if
    n_k_mx_pos=wave_par%n_mx/2
      write(0,*)'n_k_mx_pos=',n_k_mx_pos
    if(Rev2D_spectral_par_loc%max_kmx == 1) then
      n_k_mx_max=n_k_mx_pos+1
    else
!      call seperr ("max_kmx /1 is buggy")
!      n_k_mx_max=int((n_k_mx_pos)*Rev2D_spectral_par_loc%max_kmx)
      k_mx_max= &
        ((wave_par%n_mx/2+1)*wave_par%d_kmx)*Rev2D_spectral_par_loc%max_kmx
      n_k_mx_max=n_k_mx_pos+1
    end if
!      write(0,*)'k_mx_max=',k_mx_max
!      write(0,*)'n_k_mx_max=',n_k_mx_max

    if(mod(wave_par%n_z,gen_par%n_threads) /=0) then
      write(0,*)'n_threads=',gen_par%n_threads
      write(0,*)'n_z=',wave_par%n_z
      call seperr ("n_z must be multiple of n_threads")
    end if

    if(mod(n_k_mx_pos,gen_par%n_threads) /=0) then
      write(0,*)'n_threads=',gen_par%n_threads
      write(0,*)'n_k_mx_pos=',n_k_mx_pos
      call seperr ("n_k_mx_pos must be multiple of n_threads")
    end if

    if((n_k_mx_max /= n_k_mx_pos+1).and.(mod(n_k_mx_max,gen_par%n_threads) /=0)) then
      write(0,*)'n_threads=',gen_par%n_threads
      write(0,*)'n_k_mx_max=',n_k_mx_max
      call seperr ("n_k_mx_max must be multiple of n_threads")
    end if


    allocate (temp(wave_par%n_z,wave_par%n_mx), stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of temp ")
!write(0,*)'shape(temp)',shape(temp)

    allocate (temp_z(wave_par%n_z,gen_par%n_threads), stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of temp_z ")

    allocate (temp_mx(wave_par%n_mx,gen_par%n_threads), stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of temp_mx ")

!    allocate (k_z(wave_par%n_z),stat=ierr)  
!    if(ierr .ne. 0) call seperr("Problems with allocation of k_z ")
!    k_z_max=((wave_par%n_z/2+1)*wave_par%d_kz)*Rev2D_spectral_par_loc%max_kz
!    write(0,*)'k_z_max=',k_z_max
!
!    do i_k_z=1,wave_par%n_z
!       kz  = i_k_z - 1.
!       if(kz <= float(wave_par%n_z/2)) then
!          k_z(i_k_z) =  wave_par%kz_0+wave_par%d_kz* kz
!       else
!          k_z(i_k_z) = -wave_par%kz_0+wave_par%d_kz*(kz-wave_par%n_z)
!       end if
!       if(abs(k_z(i_k_z)) > k_z_max) k_z(i_k_z)=0.
!    end do
! write(0,*)'k_z=',k_z
!
!!    allocate (k_mx(n_k_mx_pos+1),stat=ierr)  
!    allocate (k_mx(wave_par%n_mx),stat=ierr)  
!    if(ierr .ne. 0) call seperr("Problems with allocation of k_mx ")
!    do i_k_mx=1,wave_par%n_mx
!       kmx  = i_k_mx - 1.
!       if(kmx <= float(wave_par%n_mx/2)) then
!          k_mx(i_k_mx) =  wave_par%kmx_0+wave_par%d_kmx* kmx
!       else
!          k_mx(i_k_mx) = -wave_par%kmx_0+wave_par%d_kmx*(kmx-wave_par%n_mx)
!       end if
!       if(abs(k_mx(i_k_mx)) > k_mx_max) k_mx(i_k_mx)=0.
!    end do
!
! write(0,*)'k_mx=',k_mx

    allocate (ksq(wave_par%n_z,wave_par%n_mx),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of ksq ")

    if(Rev2D_spectral_par_loc%max_k /= 1.) then
      limit_k=.true.
      k_max=sqrt(((wave_par%n_z/2+1)*wave_par%d_kz)**2 + &
                 ((wave_par%n_mx/2+1)*wave_par%d_kmx)**2)&
                  *Rev2D_spectral_par_loc%max_k
    else
      k_max=sqrt(((wave_par%n_z/2+2)*wave_par%d_kz)**2 + &
                 ((wave_par%n_mx/2+2)*wave_par%d_kmx)**2)
    end if
!    write(0,*)'k_max=',k_max

    do i_k_mx=1,wave_par%n_mx
       kmx  = i_k_mx - 1.
       if(kmx <= float(wave_par%n_mx/2)) then
          kk_mx =  wave_par%kmx_0+wave_par%d_kmx* kmx
       else
          kk_mx = -wave_par%kmx_0+wave_par%d_kmx*(kmx-wave_par%n_mx)
       end if
      do i_k_z=1,wave_par%n_z
         kz  = i_k_z - 0.
         if(kz <= float(wave_par%n_z/2)) then
            kk_z =  wave_par%kz_0+wave_par%d_kz* kz
         else
            kk_z = -wave_par%kz_0+wave_par%d_kz*(kz-wave_par%n_z)
         end if
         ksq(i_k_z,i_k_mx)=kk_z*kk_z + kk_mx*kk_mx
         if(sqrt(ksq(i_k_z,i_k_mx)) > k_max) ksq(i_k_z,i_k_mx)=0.
      end do
    end do

!write(0,*)'shape(ksq)',shape(ksq)
!write(0,*)'minval(sqrt(ksq))',minval(sqrt(ksq))
!write(0,*)'maxval(sqrt(ksq))',maxval(sqrt(ksq))
!! write(0,*)'ksq=',ksq

    allocate (fft_temp_z_id(gen_par%n_threads),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of fft_temp_z_id")
    if(gen_par%debug2) write(0,*)'Allocate fft_temp_z_id with shape=',shape(fft_temp_z_id)

    do i_thread=1,gen_par%n_threads
      call init_fft_1d(fft_temp_z_id(i_thread),wave_par%n_z, &
          temp_z(:,i_thread),temp_z(:,i_thread),fast_init=.false.)  
    end do

    allocate (fft_temp_mx_id(gen_par%n_threads),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of fft_temp_mx_id")
    if(gen_par%debug2) write(0,*)'Allocate fft_temp_mx_id with shape=',shape(fft_temp_mx_id)

    do i_thread=1,gen_par%n_threads
      call init_fft_1d(fft_temp_mx_id(i_thread), wave_par%n_mx, &
          temp_mx(:,i_thread),temp_mx(:,i_thread),fast_init=.false.)  
    end do

    if(gen_par%debug2) write(0,*)'scale_z=',fft_temp_z_id%scale
    if(gen_par%debug2) write(0,*)'scale_mx=',fft_temp_mx_id%scale

  end function Rev2D_spectral_init

  function Rev2D_spectral_extrap_oper( &
    rev_extrap_par, &
    wave_par, &
    force_block, &
    wave_block,slow_block) &
    result(stat)

    integer stat
    type (rev_extrap_param_type) rev_extrap_par
    type (wave_param_type) wave_par

    integer omp_get_thread_num

    logical add,adj,inject_data_loc
    integer n_z,n_mx,i_z,i_mx,i_kmx
    integer i_t0,i_tm1,i_tm2,i_z_force,half_n_z_force
!    integer i_z_force,half_n_z_force
    integer n_threads,i_thread
    real vel_sq,kmx_sq,kz_sq
    real d_t_sq
    real der2

    optional slow_block

    real,  dimension (:,:,:,:)      :: force_block
    real,  dimension (:,:,:,:,:,:)  :: wave_block
    real,  dimension (:,:,:,:,:,:)  :: slow_block

    adj=rev_extrap_par%adj
    add=rev_extrap_par%add

    n_threads=rev_extrap_par%n_threads

    d_t_sq=(rev_extrap_par%d_t)**2

    i_t0=wave_par%i_t_step(1)
    i_tm1=wave_par%i_t_step(2)
    i_tm2=wave_par%i_t_step(3)
!write(0,*)'i_t0=',i_t0
!write(0,*)'i_tm1=',i_tm1
!write(0,*)'i_tm2=',i_tm2

    n_z=wave_par%n_z
    n_mx=wave_par%n_mx

    select case (rev_extrap_par%sou_rec_wave)
     case (SOU_WAVE)
       i_z_force=i_z_sou
       half_n_z_force=half_n_z_sou

       inject_data_loc=.false.
!       write(0,*)'inject_data=',inject_data
     case (REC_WAVE)
       i_z_force=i_z_rec
       half_n_z_force=half_n_z_rec
!       write(0,*)'i_z_force=',i_z_force
!       write(0,*)'half_n_z_force=',half_n_z_force

       inject_data_loc=inject_data
!       write(0,*)'inject_data=',inject_data
       if(adj .and. limit_k) call seperr(&
            "max_k must be equal to one when propagating receivers and adj=T")
     case default
       write(0,*)'sou_rec_wave=',rev_extrap_par%sou_rec_wave
       call seperr('Rev2D_fd_mod: Not supported sou_rec_wave')
    end select


!write(0,*)'Beginning max t0 ',maxval((wave_block(i_t0,:,:,1,1,1)))
!write(0,*)'Beginning max tm1',maxval((wave_block(i_tm1,:,:,1,1,1)))
!write(0,*)'Beginning max tm2 ',maxval((wave_block(i_tm2,:,:,1,1,1)))

!write(0,*)'#1'

!$omp parallel default (private) , shared (wave_block,temp,temp_mx), &
!$omp& shared(n_z,n_mx,n_threads,i_t0,fft_temp_mx_id)
!$omp do
    do i_thread=1,n_threads
    !{
      do i_z=i_thread,n_z,n_threads
      !{
        do i_mx=1,n_mx
        !{
          temp_mx(i_mx,i_thread)=cmplx(wave_block(i_t0,i_z,i_mx,1,1,1))
        !}
        end do 
!write(0,*)'#2'
! Forward FFT along mx
        call fft_1d_data(fft_temp_mx_id(i_thread),FFT_FORW,n_mx,temp_mx(:,i_thread))
!write(0,*)'#3'
        do i_mx=1,n_mx
        !{
          temp(i_z,i_mx)=temp_mx(i_mx,i_thread)
        !}
        end do 
      !}
      end do 
    !}
    end do 
!$omp end parallel


!write(0,*)'After FFT'

!$omp parallel default (private) , shared (temp,temp_z), &
!$omp& shared(k_mx,k_z,ksq),&
!$omp& shared(n_z,n_k_mx_max,n_mx,n_threads,fft_temp_z_id)
!$omp do
    do i_thread=1,n_threads
    !{
!      do i_mx=i_thread,n_k_mx_max,n_threads
      do i_mx=i_thread,n_mx,n_threads
      !{
!        kmx_sq=k_mx(i_mx)**2
        do i_z=1,n_z
        !{
          temp_z(i_z,i_thread)=temp(i_z,i_mx)
        !}
        end do 
      
! Forward FFT along z
        call fft_1d_data(fft_temp_z_id(i_thread),FFT_FORW,n_z,temp_z(:,i_thread))

! Compute Laplacian wavefield
        do i_z=1,n_z
        !{
!          kz_sq=k_z(i_z)**2
          der2=-ksq(i_z,i_mx)
          temp_z(i_z,i_thread)=temp_z(i_z,i_thread)*der2
        !}
        end do 
! Inverse FFT along z
        call fft_1d_data(fft_temp_z_id(i_thread),FFT_BACK,n_z,temp_z(:,i_thread))
        do i_z=1,n_z
        !{
          temp(i_z,i_mx)=temp_z(i_z,i_thread)
        !}
        end do 
      !}
      end do 
    !}
    end do 
!$omp end parallel

!write(0,*)'#3'


!!$omp parallel default (private) , shared (temp), &
!!$omp& shared(n_z,n_k_mx_max,n_k_mx_pos,n_threads)
!!$omp do
!    do i_thread=1,n_threads
!    !{
!      do i_z=i_thread,n_z,n_threads
!      !{
!        do i_kmx=n_k_mx_max+1,n_k_mx_pos+1
!        !{
!          temp(i_z,i_kmx)=0
!        !}
!        end do
!      !}
!      end do 
!    !}
!    end do 
!!$omp end parallel


!! Copy into negative wavenumber using the fact that the wavefiled is real
!
!!$omp parallel default (private) , shared (temp), &
!!$omp& shared(n_z,n_mx,n_k_mx_pos,n_threads)
!!$omp do
!    do i_thread=1,n_threads
!    !{
!      do i_mx=n_k_mx_pos+1+i_thread,n_mx,n_threads
!      !{
!        do i_z=1,n_z
!        !{
!          temp(i_z,i_mx)=conjg(temp(i_z,n_mx-i_mx+2))
!        !}
!        end do 
!      !}
!      end do
!    !}
!    end do
!!$omp end parallel


!$omp parallel default (private) , shared (temp,temp_mx), &
!$omp& shared(n_z,n_mx,n_threads,fft_temp_mx_id)
!$omp do
!*$*assert do (serial)
    do i_thread=1,n_threads
    !{
      do i_z=i_thread,n_z,n_threads
      !{
        do i_mx=1,n_mx
        !{
        temp_mx(i_mx,i_thread)=temp(i_z,i_mx)
        !}
        end do 
! Inverse FFT along mx
        call fft_1d_data(fft_temp_mx_id(i_thread),FFT_BACK,n_mx,temp_mx(:,i_thread))
        do i_mx=1,n_mx
        !{
          temp(i_z,i_mx)=temp_mx(i_mx,i_thread)
        !}
        end do 
      !}
      end do 
    !}
    end do 
!$omp end parallel

!write(0,*)'#4'

! Time stepping with second order finite difference in time

!$omp parallel default (private) ,shared (wave_block,force_block,slow_block,temp),&
!$omp& shared(inject_data_loc,n_z,n_mx,n_threads,i_t0,i_tm1,i_tm2,i_z_force,half_n_z_force,d_t_sq)
!$omp do
    do i_thread=1,n_threads
    !{
      do i_mx=i_thread,n_mx,n_threads
      !{
        do i_z=1,n_z
        !{
          wave_block(i_tm2,i_z,i_mx,1,1,1)= &
              wave_block(i_tm1,i_z,i_mx,1,1,1)
          wave_block(i_tm1,i_z,i_mx,1,1,1)= &
              wave_block(i_t0,i_z,i_mx,1,1,1)
        !}
        end do

        do i_z=1,i_z_force-half_n_z_force-1
        !{
          vel_sq=1./(slow_block(i_z,i_mx,1,1,1,1)**2)
          wave_block(i_t0,i_z,i_mx,1,1,1)= &
            2*wave_block(i_tm1,i_z,i_mx,1,1,1) - &
              wave_block(i_tm2,i_z,i_mx,1,1,1) + &
              d_t_sq*(vel_sq*real(temp(i_z,i_mx)))
        !}
        end do

!write(0,*)'#5'
! Add forcing function to laplacian
        do i_z=i_z_force-half_n_z_force,i_z_force+half_n_z_force
        !{
          if(inject_data_loc) then
            wave_block(i_t0,i_z,i_mx,1,1,1)= &
                        force_block(i_z-i_z_force+half_n_z_force+1,i_mx,1,1)
          else 
            vel_sq=1./(slow_block(i_z,i_mx,1,1,1,1)**2)
            wave_block(i_t0,i_z,i_mx,1,1,1)= &
              2*wave_block(i_tm1,i_z,i_mx,1,1,1) - &
                wave_block(i_tm2,i_z,i_mx,1,1,1) + &
                d_t_sq*(vel_sq*real(temp(i_z,i_mx)) + &
                        force_block(i_z-i_z_force+half_n_z_force+1,i_mx,1,1))
          end if
        !}
        end do
!write(0,*)'#6'
  
!        wave_block(i_t0,i_ztau_sou,i_mx,1,1,1)= &
!          2*wave_block(i_tm1,i_ztau_sou,i_mx,1,1,1) - &
!            wave_block(i_tm2,i_ztau_sou,i_mx,1,1,1) + &
!            d_t_sq*(vel_sq*real(temp(i_ztau_sou,i_mx)))
!
!        wave_block(i_t0,i_ztau_sou,i_mx,1,1,1)= &
!                    wave_block(i_t0,i_ztau_sou,i_mx,1,1,1) + &
!                    force_block(i_mx,1,1,1)


        do i_z=i_z_force+half_n_z_force+1,n_z
        !{
          vel_sq=1./(slow_block(i_z,i_mx,1,1,1,1)**2)
          wave_block(i_t0,i_z,i_mx,1,1,1)= &
            2*wave_block(i_tm1,i_z,i_mx,1,1,1) - &
              wave_block(i_tm2,i_z,i_mx,1,1,1) + &
              d_t_sq*(vel_sq*real(temp(i_z,i_mx)))
        !}
        end do

!write(0,*)'#7'
      !}
      end do 
    !}
    end do 
!$omp end parallel

!write(0,*)'end prop'

     if(apply_any_bound) then
!write(0,*)'Apply absorbing boundaries'
       call apply_bound(wave_par,wave_block(i_t0,:,:,:,:,:))
     end if

!write(0,*)'max t0 ',maxval((wave_block(i_t0,:,:,1,1,1)))
!write(0,*)'max tm1',maxval((wave_block(i_tm1,:,:,1,1,1)))
!write(0,*)'max tm2 ',maxval((wave_block(i_tm2,:,:,1,1,1)))

    stat=0
    return	
  end function Rev2D_spectral_extrap_oper

  function Rev2D_spectral_scatter_oper( &
    rev_extrap_par, &
    wave_par, &
    wave_block,wave_back_block, &
    slow_block,dslow_block) &
    result(stat)

    integer stat
    type (rev_extrap_param_type) rev_extrap_par
    type (wave_param_type) wave_par

    integer omp_get_thread_num

    logical add,adj
    integer n_z,n_mx,i_z,i_mx,i_kmx
    integer i_t0
    integer n_threads,i_thread
    real d_t_sq,s_back,delta_s

    optional slow_block

    real,  dimension (:,:,:,:,:,:)  :: wave_block
    real,  dimension (:,:,:,:,:,:)  :: wave_back_block
    real,  dimension (:,:,:,:,:,:)  :: slow_block,dslow_block


    adj=rev_extrap_par%adj
    add=rev_extrap_par%add

    n_threads=rev_extrap_par%n_threads

    i_t0=wave_par%i_t_step(1)
    n_z=wave_par%n_z
    n_mx=wave_par%n_mx

    d_t_sq=(rev_extrap_par%d_t)**2

    if(.not. adj ) then
    !{
!$omp parallel default (private) ,shared (wave_block,wave_back_block,slow_block,dslow_block),&
!$omp& shared(n_z,n_mx,n_threads,i_t0)
!$omp do
      do i_thread=1,n_threads
      !{
        do i_mx=i_thread,n_mx,n_threads
        !{
          do i_z=1,n_z
          !{
            s_back=slow_block(i_z,i_mx,1,1,1,1)
            delta_s=dslow_block(i_z,i_mx,1,1,1,1)
            wave_block(i_t0,i_z,i_mx,1,1,1) = &
              wave_block(i_t0,i_z,i_mx,1,1,1) - &
              2.*s_back*delta_s*wave_back_block(1,i_z,i_mx,1,1,1)
          !}
          end do
        !}
        end do 
      !}
      end do 
!$omp end parallel
    !}
    else
    !{

!$omp parallel default (private) ,shared (wave_block,wave_back_block,slow_block,dslow_block),&
!$omp& shared(n_z,n_mx,n_threads,i_t0)
!$omp do
      do i_thread=1,n_threads
      !{
        do i_mx=i_thread,n_mx,n_threads
        !{
          do i_z=1,n_z
          !{
            s_back=slow_block(i_z,i_mx,1,1,1,1)
            dslow_block(i_z,i_mx,1,1,1,1)= &
              dslow_block(i_z,i_mx,1,1,1,1) - &
                2.*s_back*wave_back_block(1,i_z,i_mx,1,1,1)* &
                wave_block(i_t0,i_z,i_mx,1,1,1)
          !}
          end do
        !}
        end do 
      !}
      end do 
!$omp end parallel
    !}
    end if

    stat=0
    return	
  end function Rev2D_spectral_scatter_oper

  function Rev2D_spectral_der2time_oper (&
    rev_extrap_par, &
    wave_par, &
    wave_block, &
    wave_back_block) &
    result(stat)

    integer stat
    type (rev_extrap_param_type) rev_extrap_par
    type (wave_param_type) wave_par

    integer omp_get_thread_num

    logical add,adj
    integer n_z,n_mx,i_z,i_mx
    integer i_t0,i_tm1,i_tm2
    integer n_threads,i_thread

    real,  dimension (:,:,:,:,:,:)  :: wave_block
    real,  dimension (:,:,:,:,:,:)  :: wave_back_block

    adj=rev_extrap_par%adj
    add=rev_extrap_par%add

    n_threads=rev_extrap_par%n_threads

    i_t0=wave_par%i_t_step(1)
    i_tm1=wave_par%i_t_step(2)
    i_tm2=wave_par%i_t_step(3)

    n_z=wave_par%n_z
    n_mx=wave_par%n_mx

! second derivative in time 

!$omp parallel default (private) ,shared (wave_block,wave_back_block),&
!$omp& shared(n_z,n_mx,n_threads,i_t0,i_tm1,i_tm2)
!$omp do
    do i_thread=1,n_threads
    !{
      do i_mx=i_thread,n_mx,n_threads
      !{
        do i_z=1,n_z
        !{
          wave_back_block(1,i_z,i_mx,1,1,1)= &
              wave_block(i_t0,i_z,i_mx,1,1,1) - &
            2*wave_block(i_tm1,i_z,i_mx,1,1,1) + &
              wave_block(i_tm2,i_z,i_mx,1,1,1) 
        !}
        end do
      !}
      end do 
    !}
    end do 
!$omp end parallel

    stat=0
    return	
  end function Rev2D_spectral_der2time_oper

  function Rev2D_spectral_clean()  result(stat)
    integer stat

    integer ierr

    if(allocated(temp)) then
       deallocate (temp,stat=ierr)  
       if(ierr /= 0) call seperr('Rev2D_spectral_clean: Problems with deallocating temp')
    end if
    if(allocated(temp_z)) then
       deallocate (temp_z,stat=ierr)  
       if(ierr /= 0) call seperr('Rev2D_spectral_clean: Problems with deallocating temp_z')
    end if
    if(allocated(temp_mx)) then
       deallocate (temp_mx,stat=ierr)  
       if(ierr /= 0) call seperr('Rev2D_spectral_clean: Problems with deallocating temp_mx')
    end if
!    if(allocated(k_z)) then
!       deallocate (k_z,stat=ierr)  
!       if(ierr /= 0) call seperr('Rev2D_spectral_clean: Problems with deallocating k_z')
!    end if
!    if(allocated(k_mx)) then
!       deallocate (k_mx,stat=ierr)  
!       if(ierr /= 0) call seperr('Rev2D_spectral_clean: Problems with deallocating k_mx')
!    end if
    if(allocated(ksq)) then
       deallocate (ksq,stat=ierr)  
       if(ierr /= 0) call seperr('Rev2D_spectral_clean: Problems with deallocating ksq')
    end if
    if(associated(fft_temp_z_id)) then
       deallocate (fft_temp_z_id,stat=ierr)  
       if(ierr /= 0) call seperr('Rev2D_spectral_clean: Problems with deallocating fft_temp_z_id')
    end if
    if(associated(fft_temp_mx_id)) then
       deallocate (fft_temp_mx_id,stat=ierr)  
       if(ierr /= 0) call seperr('Rev2D_spectral_clean: Problems with deallocating fft_temp_mx_id')
    end if

    stat=0

       write(0,*)' Rev2D_spectral_clean: Done with cleaning'

  end function Rev2D_spectral_clean


end module Rev2D_spectral_mod
