module CorrMov_mod

  use CorrMov_types

  implicit none

  integer, private :: n_bytes_maxima,n_bytes_data,n_bytes_stack
  integer, private :: n_bytes_r_weights,n_bytes_l_weights
  integer, private :: n_bytes_beam
  real, private :: norm
  integer, external ::  sreed,srite,sseek_block

!  integer, allocatable, dimension (:), private :: max_loc
  real, allocatable, dimension (:,:), private :: maxima,l_weights
  real, allocatable, dimension (:), private :: r_weights,stack
  real, allocatable, dimension (:,:), private :: l_weights_eff
  real, allocatable, dimension (:,:), private :: data_in,data_out
  real, allocatable, dimension (:,:,:), private :: data_beam
!  complex, allocatable, dimension (:), private :: coord

contains

function CorrMov_init(CorrMov_par)  result(status)
  type (CorrMov_param_type) CorrMov_par
  integer status

  integer ierr

  n_bytes_data=CorrMov_par%n1*CorrMov_par%n2*CorrMov_par%esize
  n_bytes_maxima=CorrMov_par%n1_max*CorrMov_par%n2_max*CorrMov_par%esize
  n_bytes_stack=CorrMov_par%n1*CorrMov_par%esize
  n_bytes_r_weights=CorrMov_par%n2*CorrMov_par%esize
  n_bytes_beam=CorrMov_par%n1*(1+2*CorrMov_par%h_width_bform)* &
               CorrMov_par%n2*CorrMov_par%esize
write(0,*)'n_bytes_beam=',n_bytes_beam


  allocate (data_in(CorrMov_par%n1,CorrMov_par%n2),stat=ierr)  
  if(ierr .ne. 0) call seperr("Problems with allocation of data_in")

  select case (CorrMov_par%oper_type(1:3))
  case ('pic')
    allocate (maxima(CorrMov_par%n1_max,CorrMov_par%n2_max),stat=ierr)  
    if(ierr .ne. 0) call seperr("Problems with allocation of maxima")
  case ('dso')
    allocate (data_out(CorrMov_par%n1,CorrMov_par%n2),stat=ierr) 
    if(ierr .ne. 0) call seperr("Problems with allocation of data_out")
    if(CorrMov_par%r_weights_dso) then
      allocate (r_weights(CorrMov_par%n2),stat=ierr) 
      if(ierr .ne. 0) call seperr("Problems with allocation of r_weights")
    end if
    if(CorrMov_par%l_weights_dso) then
      allocate (l_weights(CorrMov_par%n2,1),stat=ierr) 
      if(ierr .ne. 0) call seperr("Problems with allocation of l_weights")
      n_bytes_l_weights=CorrMov_par%n2*CorrMov_par%esize
    end if
  case ('sta')
    allocate (stack(CorrMov_par%n1),stat=ierr) 
    if(ierr .ne. 0) call seperr("Problems with allocation of stack")
    
    if(CorrMov_par%norm_stack) then
      if(CorrMov_par%unitary_stack) then
        norm=sqrt(1./CorrMov_par%n2)
      else
        if(CorrMov_par%first_adj) then
          norm=1.
        else
          norm=(1./CorrMov_par%n2)
        end if
      end if
    else
      norm=1.
    end if

    if(CorrMov_par%r_weights_stack) then
      allocate (r_weights(CorrMov_par%n2),stat=ierr) 
      if(ierr .ne. 0) call seperr("Problems with allocation of r_weights")
    end if

  case ('l-s')
    allocate (data_out(CorrMov_par%n1,CorrMov_par%n2),stat=ierr) 
    if(ierr .ne. 0) call seperr("Problems with allocation of data_out")
    if(CorrMov_par%r_weights_lstack) then
      allocate (r_weights(CorrMov_par%n2),stat=ierr) 
      if(ierr .ne. 0) call seperr("Problems with allocation of r_weights")
    end if
    if(CorrMov_par%l_weights_lstack) then
      allocate (l_weights(CorrMov_par%n2,1),stat=ierr) 
      if(ierr .ne. 0) call seperr("Problems with allocation of l_weights")
      n_bytes_l_weights=CorrMov_par%n2*CorrMov_par%esize
    end if

    if(CorrMov_par%norm_lstack) then
      if(CorrMov_par%unitary_lstack) then
        norm=sqrt(1./3)
      else
        if(CorrMov_par%first_adj) then
          norm=1.
        else
          norm=(1./3)
        end if
      end if
    else
      norm=1.
    end if

  case ('b-f')
    allocate (data_beam(CorrMov_par%n1, &
                        -CorrMov_par%h_width_bform:CorrMov_par%h_width_bform, &
                        CorrMov_par%n2),stat=ierr) 
    if(ierr .ne. 0) call seperr("Problems with allocation of data_beam")
    if(CorrMov_par%r_weights_bform) then
      allocate (r_weights(CorrMov_par%n2),stat=ierr) 
      if(ierr .ne. 0) call seperr("Problems with allocation of r_weights")
    end if
    if(CorrMov_par%l_weights_bform) then
      allocate (l_weights( &
        -CorrMov_par%h_width_bform:CorrMov_par%h_width_bform,CorrMov_par%n2) &
         ,stat=ierr) 
      allocate (l_weights_eff( &
        -CorrMov_par%h_width_bform:CorrMov_par%h_width_bform,CorrMov_par%n2) &
         ,stat=ierr) 
      n_bytes_l_weights=(1+2*CorrMov_par%h_width_bform)*CorrMov_par%n2* &
                         CorrMov_par%esize
!write(0,*)'n_bytes_l_weights=',n_bytes_l_weights
      if(ierr .ne. 0) call seperr("Problems with allocation of l_weights")
    end if

  case default
     write(0,*)'CorrMov_par-oper_type=',trim(CorrMov_par%oper_type)
     call seperr('Not supported oper_type')
  end select


  status=ierr

end function CorrMov_init

function CorrMov_oper(adj,add,CorrMov_par)  result(status)
  logical adj,add
  type (CorrMov_param_type) CorrMov_par
  integer status

  integer i1,i2,i3,i4,i5,max_loc,i2m1,i2p1,pos,neg
  integer i1_min,i1_max,i2_min,i2_max,i2_sum,i2_beam
  integer i2_origin,i2_sum_beg,i2_sum_end,i2_beam_beg,i2_beam_end,i_abs2
  integer h_w_left_pos,h_w_right_pos,h_w_left_neg,h_w_right_neg
  integer nsrite
  real max_val,min_pick,min_dso,max_r_weights,min_lstack
  real beam_w,abs2,min_abs2,max_abs2,scale_l_w


  select case (CorrMov_par%oper_type(1:3))
  case ('pic')
  !{
 
    do i5=1,CorrMov_par%n5
    !{ i5
      do i4=1,CorrMov_par%n4
      !{ i4
  
        do i3=1,CorrMov_par%n3
        !{ i3
          if(sreed('in',data_in,n_bytes_data) /= n_bytes_data) then
            call seperr('Problems with reading input data')
          end if
  
          if(CorrMov_par%pick_threshold > 0.) then
            min_pick=CorrMov_par%pick_threshold*maxval(data_in)
          else
            min_pick=-huge(max_val)
          end if

          do i2=1,CorrMov_par%n2
          !{ i2
            max_val=min_pick
            max_loc=-1
!write(0,*)'max_val init=',max_val
            do i1=1,CorrMov_par%n1
            !{ i1
              if(data_in(i1,i2) > max_val) then
                max_val=data_in(i1,i2)
                max_loc=i1
              end if
            !} i1
            end do
!write(0,*)'max_val end=',max_val
!write(0,*)'max_loc end=',max_loc
            if(max_loc >0) then
              maxima(i2,i3)=CorrMov_par%o1+(max_loc-1)*CorrMov_par%d1
            else
              maxima(i2,i3)=0.
            end if
          !} i2
          end do
        !} i3
        end do
  
        if(srite('out',maxima,n_bytes_maxima) /= n_bytes_maxima) then
          call seperr('Problems with writing Picked Maxima')
        end if
  
      !} i4
      end do
    !} i5
    end do
  !}
  case ('dso')
  !{
    if(adj) then 
      pos=-1
!      pos=0
      neg=1
    else
      pos=1
!      pos=0
      neg=-1
    end if
               
    i1_min=nint((CorrMov_par%dso_min1-CorrMov_par%o1)/CorrMov_par%d1)+1
    i1_max=nint((CorrMov_par%dso_max1-CorrMov_par%o1)/CorrMov_par%d1)+1
    i2_min=nint((CorrMov_par%dso_min2-CorrMov_par%o2)/CorrMov_par%d2)+1
    i2_max=nint((CorrMov_par%dso_max2-CorrMov_par%o2)/CorrMov_par%d2)+1
!write(0,*)'i1_min=',i1_min
!write(0,*)'i1_max=',i1_max
    do i5=1,CorrMov_par%n5
    !{ i5
      do i4=1,CorrMov_par%n4
      !{ i4
        do i3=1,CorrMov_par%n3
        !{
          data_out=0
          if(sreed('in',data_in,n_bytes_data) /= n_bytes_data) then
            call seperr('Problems with reading input data')
          end if
          if(CorrMov_par%test) then
             data_in=1.
          end if
          if(CorrMov_par%r_weights_dso) then
          !{
!write(0,*)'Read R_Weight'
            r_weights=0
            if(sreed(trim(CorrMov_par%tag_r_weights),r_weights,n_bytes_r_weights)&
                /= n_bytes_r_weights) then
              call seperr('Problems with reading input R_Weights')
            end if
            max_r_weights=maxval(r_weights(:))
            do i2=1,CorrMov_par%n2
              if(r_weights(i2) >=  CorrMov_par%dso_epsi*max_r_weights) then
                r_weights(i2)=1./r_weights(i2)
              else
                r_weights(i2)=1./(CorrMov_par%dso_epsi*max_r_weights)
!                r_weights(i2,i3)=1.
              end if
            end do

            if(.not. adj) then 
  !            r_weights=1.
              do i1=i1_min,i1_max
                data_in(i1,:)=data_in(i1,:)*r_weights(:)
  !              data_in(i1,:,:)=r_weights(:,:)
              end do
            end if
          !}
          end if
          if(CorrMov_par%l_weights_dso) then
          !{
!write(0,*)'Read L_Weight'
            l_weights=0
            if(sreed(trim(CorrMov_par%tag_l_weights),l_weights,n_bytes_l_weights)&
                /= n_bytes_l_weights) then
              call seperr('Problems with reading input L_Weights')
            end if
            if(adj) then
!write(0,*)'Apply L_Weight with Adj=y'
!write(0,*)'minval(l_weights)=',minval(l_weights)
!write(0,*)'maxval(l_weights)=',maxval(l_weights)
              do i1=i1_min,i1_max
                data_in(i1,:)=data_in(i1,:)*l_weights(:,1)
              end do
            end if
          !}
          end if

          if(CorrMov_par%dso_threshold > 0.) then
            min_dso=CorrMov_par%dso_threshold*maxval(data_in)
          else
            min_dso=-huge(max_val)
          end if
          if(adj) then 
          !{
            if(i2_min ==1) then
              do i1=i1_min,i1_max
              !{ i1
                data_out(i1,1)= &
                  data_in(i1,2)-data_in(i1,1)
              !} i1
              end do
            end if
            if(i2_max == CorrMov_par%n2) then
              do i1=i1_min,i1_max
              !{ i1
                data_out(i1,CorrMov_par%n2)=& 
                  data_in(i1,CorrMov_par%n2)-data_in(i1,CorrMov_par%n2-1)
              !} i1
              end do
            end if
          !}
          else
          !{
            if(i2_min ==1) then
              do i1=i1_min,i1_max
              !{ i1
                data_out(i1,1)= &
                  -data_in(i1,2)-data_in(i1,1)
              !} i1
              end do
            end if
            if(i2_max == CorrMov_par%n2) then
              do i1=i1_min,i1_max
              !{ i1
                data_out(i1,CorrMov_par%n2)=& 
                  data_in(i1,CorrMov_par%n2)+data_in(i1,CorrMov_par%n2-1)
              !} i1
              end do
            end if
          !}
          end if

          do i2=max(2,i2_min+1),min(CorrMov_par%n2-1,i2_max-1)
          !{ i2
            i2m1=i2+neg
            i2p1=i2+pos
            do i1=i1_min,i1_max
            !{ i1
              if(min(abs(data_in(i1,i2p1)),abs(data_in(i1,i2m1))) > & 
                      min_dso) then
                data_out(i1,i2)=data_in(i1,i2p1)-data_in(i1,i2m1)
              else
                data_out(i1,i2)=0.
              end if
            !} i1
            end do
          !} i2
          end do

          if(CorrMov_par%r_weights_dso .and. adj) then
            do i1=i1_min,i1_max
              data_out(i1,:)=data_out(i1,:)*r_weights(:)
            end do
          end if
          if(CorrMov_par%l_weights_dso .and. (.not. adj)) then
!write(0,*)'Apply L_Weight with Adj=n'
!write(0,*)'maxval(l_weights)=',maxval(l_weights)
            do i1=i1_min,i1_max
              data_out(i1,:)=data_out(i1,:)*l_weights(:,1)
            end do
          end if
          if(srite('out',data_out,n_bytes_data) /= n_bytes_data) then
            call seperr('Problems with writing DSO output')
          end if
        !} i3
        end do
      !} i4
      end do
    !} i5
    end do
  !}
  case ('sta')
  !{
    do i5=1,CorrMov_par%n5
    !{ i5
      do i4=1,CorrMov_par%n4
      !{ i4
        do i3=1,CorrMov_par%n3
        !{
          if(CorrMov_par%r_weights_stack) then
          !{
!write(0,*)'Read R_Weight'
            r_weights=0
            if(sreed(trim(CorrMov_par%tag_r_weights),r_weights,n_bytes_r_weights)&
                /= n_bytes_r_weights) then
              call seperr('Problems with reading input R_Weights')
            end if
            max_r_weights=maxval(r_weights(:))
!      write(0,*)'Max w',max_r_weights
!      write(0,*)'Min w',1./(CorrMov_par%stack_epsi*max_r_weights)
            do i2=1,CorrMov_par%n2
              if(r_weights(i2) >=  CorrMov_par%stack_epsi*max_r_weights) then
                r_weights(i2)=1./r_weights(i2)
              else
                r_weights(i2)=1./(CorrMov_par%stack_epsi*max_r_weights)
              end if
            end do
! call snap('R_W.H',CorrMov_par%n2,CorrMov_par%n3,r_weights)

          if(.not. adj) then 
!            r_weights=1.
            do i1=i1_min,i1_max
              data_in(i1,:)=data_in(i1,:)*r_weights(:)
!              data_in(i1,:,:)=r_weights(:,:)
            end do
          end if
        !}
        end if

        if(adj) then 
        !{ 
          if(sreed('in',stack,n_bytes_stack) /= n_bytes_stack) then
            call seperr('Problems with reading input stack')
          end if
          stack= norm*stack
          do i2=1,CorrMov_par%n2
          !{ i2
            do i1=1,CorrMov_par%n1
            !{ i1
              data_in(i1,i2)=stack(i1)
            !} i1
            end do
          !} i2
          end do
          if(CorrMov_par%r_weights_stack) then
            do i1=1,CorrMov_par%n1
              data_in(i1,:)=data_in(i1,:)*r_weights(:)
            end do
          end if
          if(srite('out',data_in,n_bytes_data) /= n_bytes_data) then
            call seperr('Problems with writing data output')
          end if
          !} 
          else
          !{ 
            if(sreed('in',data_in,n_bytes_data) /= n_bytes_data) then
              call seperr('Problems with reading input data')
            end if
            if(CorrMov_par%r_weights_stack) then
              do i1=1,CorrMov_par%n1
                data_in(i1,:)=data_in(i1,:)*r_weights(:)
              end do
            end if
            stack=0
            do i2=1,CorrMov_par%n2
            !{ i2
              do i1=1,CorrMov_par%n1
              !{ i1
                stack(i1)= stack(i1)+data_in(i1,i2)
              !} i1
              end do
            !} i2
            end do
            stack= norm*stack
            if(srite('out',stack,n_bytes_stack) /= n_bytes_stack) then
              call seperr('Problems with writing Stack output')
            end if
          !} 
          end if
        !} i3
        end do
      !} i4
      end do
    !} i5
    end do
  !}

  case ('l-s')
  !{
    i1_min=nint((CorrMov_par%lstack_min1-CorrMov_par%o1)/CorrMov_par%d1)+1
    i1_max=nint((CorrMov_par%lstack_max1-CorrMov_par%o1)/CorrMov_par%d1)+1
    i2_min=nint((CorrMov_par%lstack_min2-CorrMov_par%o2)/CorrMov_par%d2)+1
    i2_max=nint((CorrMov_par%lstack_max2-CorrMov_par%o2)/CorrMov_par%d2)+1
!write(0,*)'i1_min=',i1_min
!write(0,*)'i1_max=',i1_max
    if(CorrMov_par%asym_lstack) then
     i2_origin=-nint(CorrMov_par%o2/CorrMov_par%d2)+1
write(0,*)'i2_origin=',i2_origin
       if(adj) then 
         h_w_left_neg=0
         h_w_right_neg=CorrMov_par%h_width_lstack
         h_w_left_pos=CorrMov_par%h_width_lstack
         h_w_right_pos=0
       else 
         h_w_left_neg=CorrMov_par%h_width_lstack
         h_w_right_neg=0
         h_w_left_pos=0
         h_w_right_pos=CorrMov_par%h_width_lstack
       end if
    else
      i2_origin=i2_min
      h_w_left_neg=CorrMov_par%h_width_lstack
      h_w_right_neg=CorrMov_par%h_width_lstack
      h_w_left_pos=CorrMov_par%h_width_lstack
      h_w_right_pos=CorrMov_par%h_width_lstack
    end if
    do i5=1,CorrMov_par%n5
    !{ i5
      do i4=1,CorrMov_par%n4
      !{ i4
        do i3=1,CorrMov_par%n3
        !{
          data_out=0
          if(sreed('in',data_in,n_bytes_data) /= n_bytes_data) then
            call seperr('Problems with reading input data')
          end if
          if(CorrMov_par%test) then
             data_in=1.
          end if
          if(CorrMov_par%r_weights_lstack) then
          !{
!write(0,*)'Read R_Weight'
            r_weights=0
            if(sreed(trim(CorrMov_par%tag_r_weights),r_weights,n_bytes_r_weights)&
                /= n_bytes_r_weights) then
              call seperr('Problems with reading input R_Weights')
            end if
            max_r_weights=maxval(r_weights(:))
!write(0,*)'Apply R_Weight with Adj=y'
!write(0,*)'minval(r_weights)=',minval(r_weights)
!write(0,*)'maxval(r_weights)=',maxval(r_weights)
            do i2=1,CorrMov_par%n2
              if(r_weights(i2) >= CorrMov_par%lstack_epsi*max_r_weights) then
                r_weights(i2)=1./r_weights(i2)
              else
                r_weights(i2)=1./(CorrMov_par%lstack_epsi*max_r_weights)
!                r_weights(i2,i3)=1.
              end if
            end do

            if(.not. adj) then 
!            r_weights=1.
              do i1=i1_min,i1_max
                data_in(i1,:)=data_in(i1,:)*r_weights(:)
!              data_in(i1,:,:)=r_weights(:,:)
              end do
            end if
          !}
          end if
          if(CorrMov_par%l_weights_lstack) then
          !{
!write(0,*)'Read L_Weight'
            l_weights=0
            if(sreed(trim(CorrMov_par%tag_l_weights),l_weights,n_bytes_l_weights)&
                /= n_bytes_l_weights) then
              call seperr('Problems with reading input L_Weights')
            end if
            if(adj) then
  !write(0,*)'Apply L_Weight with Adj=y'
  !write(0,*)'minval(l_weights)=',minval(l_weights)
  !write(0,*)'maxval(l_weights)=',maxval(l_weights)
              do i1=i1_min,i1_max
                data_in(i1,:)=data_in(i1,:)*l_weights(:,1)
              end do
            end if
          !}
          end if

          if(CorrMov_par%lstack_threshold > 0.) then
            min_lstack=CorrMov_par%lstack_threshold*maxval(data_in)
          else
            min_lstack=-huge(max_val)
          end if
          do i2=max(1,i2_min),min(CorrMov_par%n2,i2_max)
          !{ i2
            data_out(i1_min:i1_max,i2)=0
            if (i2 < i2_origin) then
              i2_sum_beg=max(-h_w_left_neg,1-i2)
              i2_sum_end=min(h_w_right_neg,CorrMov_par%n2-i2)
            else
              i2_sum_beg=max(-h_w_left_pos,1-i2)
              i2_sum_end=min(h_w_right_pos,CorrMov_par%n2-i2)
            end if
            do i2_sum=i2_sum_beg,i2_sum_end
!            do i2_sum=max(-CorrMov_par%h_width_lstack,1-i2), &
!                      min(CorrMov_par%h_width_lstack,CorrMov_par%n2-i2)
            !{ i2_sum
              do i1=i1_min,i1_max
              !{ i1
                data_out(i1,i2)= &
                  data_out(i1,i2)+data_in(i1,i2+i2_sum)
              !} i1
              end do
            !} i2_sum
            end do
          !} i2
          end do

          if(CorrMov_par%r_weights_lstack .and. adj) then
            do i1=i1_min,i1_max
              data_out(i1,:)=data_out(i1,:)*r_weights(:)
            end do
          end if
          if(CorrMov_par%l_weights_lstack .and. (.not. adj)) then
            do i1=i1_min,i1_max
              data_out(i1,:)=data_out(i1,:)*l_weights(:,1)
            end do
          end if
          if(srite('out',data_out,n_bytes_data) /= n_bytes_data) then
            call seperr('Problems with writing Local Stacks output')
          end if
        !} i3
        end do
      !} i4
      end do
    !} i5
    end do
  !}

  case ('b-f')
  !{
    i1_min=nint((CorrMov_par%bform_min1-CorrMov_par%o1)/CorrMov_par%d1)+1
    i1_max=nint((CorrMov_par%bform_max1-CorrMov_par%o1)/CorrMov_par%d1)+1
    i2_min=nint((CorrMov_par%bform_min2-CorrMov_par%o2)/CorrMov_par%d2)+1
    i2_max=nint((CorrMov_par%bform_max2-CorrMov_par%o2)/CorrMov_par%d2)+1
!write(0,*)'i2_min=',i2_min
!write(0,*)'i2_max=',i2_max
    do i5=1,CorrMov_par%n5
    !{ i5
      do i4=1,CorrMov_par%n4
      !{ i4
        do i3=1,CorrMov_par%n3
        !{
          if(adj) then 
            data_in=0
            if(sreed('in',data_beam,n_bytes_beam) /= n_bytes_beam) then
              call seperr('Problems with reading input beam formed data')
            end if
          else
            data_beam=0
            if(sreed('in',data_in,n_bytes_data) /= n_bytes_data) then
              call seperr('Problems with reading input data')
            end if
          end if
          if(CorrMov_par%test) then
             call seperr('No test for b-f')
          end if
          if(CorrMov_par%r_weights_bform) then
          !{
!write(0,*)'Read R_Weight'
            r_weights=0
            if(sreed(trim(CorrMov_par%tag_r_weights),r_weights,n_bytes_r_weights)&
                /= n_bytes_r_weights) then
              call seperr('Problems with reading input R_Weights')
            end if

            max_r_weights=maxval(r_weights(:))
!write(0,*)'Normalize R_Weight with Adj=y'
!write(0,*)'minval(r_weights)=',minval(r_weights)
!write(0,*)'maxval(r_weights)=',maxval(r_weights)
            do i2=1,CorrMov_par%n2
              if(r_weights(i2) >= CorrMov_par%bform_epsi*max_r_weights) then
                r_weights(i2)=1./r_weights(i2)
              else
                r_weights(i2)=1./(CorrMov_par%bform_epsi*max_r_weights)
!                r_weights(i2,i3)=1.
              end if
            end do
            if(.not. adj) then 
!write(0,*)'Apply R_Weight with Adj=y'
              do i1=i1_min,i1_max
                data_in(i1,:)=data_in(i1,:)*r_weights(:)
              end do
            end if
          !}
          end if

          if(CorrMov_par%l_weights_bform) then
          !{
write(0,*)'Read L_Weight'
            l_weights=0
            if(sreed(trim(CorrMov_par%tag_l_weights),l_weights, &
                n_bytes_l_weights)/= n_bytes_l_weights) then
              call seperr('Problems with reading input L_Weights')
            end if
            do i2=max(1,i2_min),min(CorrMov_par%n2,i2_max)
            !{ i2
              i2_beam_beg=max(-CorrMov_par%h_width_bform,1-i2)
              i2_beam_end=min(CorrMov_par%h_width_bform,CorrMov_par%n2-i2)
              min_abs2=huge(min_abs2)
              max_abs2=tiny(max_abs2)
              do i2_beam=i2_beam_beg,i2_beam_end
              !{ i2_beam
                 abs2=abs((CorrMov_par%o2+(i2-1)*CorrMov_par%d2) + &
                           i2_beam*CorrMov_par%d2)
                 min_abs2=min(min_abs2,abs2)
                 max_abs2=max(max_abs2,abs2)
              !} i2_beam
              end do
              scale_l_w=(((max_abs2-min_abs2)/CorrMov_par%d2)+1)/ &
                        (1+2*CorrMov_par%h_width_bform)
!write(0,*)'**'
!write(0,*)'i2=',i2,'min_abs2=',min_abs2,'max_abs2=',max_abs2
!write(0,*)'scale_l_w=',scale_l_w
!!write(0,*)'i2_beam_beg=',i2_beam_beg,'i2_beam_end=',i2_beam_end
!write(0,*)'**'
              do i2_beam=i2_beam_beg,i2_beam_end
              !{ i2_beam
                 abs2=abs((CorrMov_par%o2+(i2-1)*CorrMov_par%d2) + &
                           i2_beam*CorrMov_par%d2)
                 i_abs2=nint((((abs2-min_abs2)/CorrMov_par%d2)/scale_l_w) - &
                                CorrMov_par%h_width_bform)
                 i_abs2=max(-CorrMov_par%h_width_bform, &
                            min(CorrMov_par%h_width_bform,i_abs2))
                 l_weights_eff(i2_beam,i2)=l_weights(i_abs2,i2)
              !} i2_beam
              end do
            !} i2
            end do
 call snap('L_W.H',1+2*CorrMov_par%h_width_bform,CorrMov_par%n2,l_weights_eff)
            if(adj) then
write(0,*)'Apply L_Weight with Adj=y'
!write(0,*)'minval(l_weights)=',minval(l_weights)
!write(0,*)'maxval(l_weights)=',maxval(l_weights)
              do i1=i1_min,i1_max
                data_beam(i1,:,:)=data_beam(i1,:,:)*l_weights_eff(:,:)
              end do
            end if
          !}
          end if

!write(0,*)'i3=',i3,'i2_min=',i2_min,'i2_max=',i2_max
          do i2=max(1,i2_min),min(CorrMov_par%n2,i2_max)
          !{ i2
            i2_beam_beg=max(-CorrMov_par%h_width_bform,1-i2)
            i2_beam_end=min(CorrMov_par%h_width_bform,CorrMov_par%n2-i2)
            do i2_beam=i2_beam_beg,i2_beam_end
            !{ i2_beam
              if(CorrMov_par%beam_weight_power /= 0.) then
                beam_w=(i2_beam*CorrMov_par%d2)**CorrMov_par%beam_weight_power
              else
                beam_w=1.
              end if
              if(adj) then 
                do i1=i1_min,i1_max
                  data_in(i1,i2+i2_beam)= data_in(i1,i2+i2_beam) + &
                                             data_beam(i1,i2_beam,i2)*beam_w
                end do
              else
                do i1=i1_min,i1_max
                  data_beam(i1,i2_beam,i2)= data_in(i1,i2+i2_beam)*beam_w
                end do
              end if
            !} i2_beam
            end do
          !} i2
          end do

          if(CorrMov_par%l_weights_bform .and. (.not. adj)) then
write(0,*)'Apply L_Weight with Adj=n'
            do i1=i1_min,i1_max
              data_beam(i1,:,:)=data_beam(i1,:,:)*l_weights_eff(:,:)
            end do
          end if

          if(adj) then
            if(CorrMov_par%r_weights_bform) then
!write(0,*)'Apply R_Weight with Adj=y'
              do i1=i1_min,i1_max
                data_in(i1,:)=data_in(i1,:)*r_weights(:)
              end do
            end if
!write(0,*)'Writing with Adj=y'
            if(srite('out',data_in,n_bytes_data) /= n_bytes_data) then
              call seperr('Problems with writing Beam Recomposed output')
            end if
          else
!write(0,*)'Writing with Adj=n'
!write(0,*)'n_bytes_beam=',n_bytes_beam
!          nsrite=srite('out',data_beam,n_bytes_beam) 
!write(0,*)'nsrite=',nsrite
          if(srite('out',data_beam,n_bytes_beam) /= n_bytes_beam) then
            call seperr('Problems with writing Beam Forming output')
          end if
          end if
        !} i3
        end do
      !} i4
      end do
    !} i5
    end do
  !}

  case default
     write(0,*)'CorrMov_par-oper_type=',trim(CorrMov_par%oper_type)
     call seperr('Not supported oper_type')
  end select
   
  status=0

end function CorrMov_oper

function CorrMov_clean()  result(stat)
    integer stat

    integer ierr

!    if(allocated(max_loc)) then
!       deallocate (max_loc,stat=ierr)  
!       if(ierr /= 0) call seperr('Problems with deallocating max_loc')
!    end if
    if(allocated(maxima)) then
       deallocate (maxima,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating maxima')
    end if
    if(allocated(data_in)) then
       deallocate (data_in,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating data_in')
    end if
    if(allocated(data_out)) then
       deallocate (data_out,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating data_out')
    end if
    if(allocated(data_beam)) then
       deallocate (data_beam,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating data_beam')
    end if
    if(allocated(stack)) then
       deallocate (stack,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating stack')
    end if
    if(allocated(r_weights)) then
       deallocate (r_weights,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating r_weights')
    end if
    if(allocated(l_weights)) then
       deallocate (l_weights,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating l_weights')
    end if
    if(allocated(l_weights_eff)) then
       deallocate (l_weights_eff,stat=ierr)  
       if(ierr /= 0) call seperr('Problems with deallocating l_weights_eff')
    end if



    stat=ierr

end function CorrMov_clean

end module CorrMov_mod
